/*
** Copyright (C) 2002-2004 Daniel Caujolle-Bert <segfault@club-internet.fr>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <pthread.h>

#include <readline.h>
#include <tilde.h>
#include <history.h>

#include "common.h"
#include "utils.h"
#include "parse.h"
#include "commands.h"
#include "xine_commands.h"
#include "loader.h"
#include "script.h"
#include "playlist.h"

extern int errno;

static struct event_struct_s {
  char *name;
  uint32_t event;
} event_struct[] = {
  { "UI_PLAYBACK_FINISHED", XINE_EVENT_UI_PLAYBACK_FINISHED    },
  { "UI_CHANNELS_CHANGED" , XINE_EVENT_UI_CHANNELS_CHANGED     },
  { "UI_SET_TITLE"        , XINE_EVENT_UI_SET_TITLE            },
  { "UI_MESSAGE"          , XINE_EVENT_UI_MESSAGE              },
  { "FRAME_FORMAT_CHANGE" , XINE_EVENT_FRAME_FORMAT_CHANGE     },
  { "AUDIO_LEVEL"         , XINE_EVENT_AUDIO_LEVEL             },
  { "QUIT"                , XINE_EVENT_QUIT                    },
  { "PROGRESS"            , XINE_EVENT_PROGRESS                },
  { "MRL_REFERENCE"       , XINE_EVENT_MRL_REFERENCE           },
  { "UI_NUM_BUTTONS"      , XINE_EVENT_UI_NUM_BUTTONS          },
  { "SPU_BUTTON"          , XINE_EVENT_SPU_BUTTON              },
  { "DROPPED_FRAMES"      , XINE_EVENT_DROPPED_FRAMES          },
  { "INPUT_MOUSE_BUTTON"  , XINE_EVENT_INPUT_MOUSE_BUTTON      }, 
  { "INPUT_MOUSE_MOVE"    , XINE_EVENT_INPUT_MOUSE_MOVE        },
  { "INPUT_MENU1"         , XINE_EVENT_INPUT_MENU1             },
  { "INPUT_MENU2"         , XINE_EVENT_INPUT_MENU2             },
  { "INPUT_MENU3"         , XINE_EVENT_INPUT_MENU3             },
  { "INPUT_MENU4"         , XINE_EVENT_INPUT_MENU4             },
  { "INPUT_MENU5"         , XINE_EVENT_INPUT_MENU5             },
  { "INPUT_MENU6"         , XINE_EVENT_INPUT_MENU6             },
  { "INPUT_MENU7"         , XINE_EVENT_INPUT_MENU7             },
  { "INPUT_UP"            , XINE_EVENT_INPUT_UP                },
  { "INPUT_DOWN"          , XINE_EVENT_INPUT_DOWN              },
  { "INPUT_LEFT"          , XINE_EVENT_INPUT_LEFT              },
  { "INPUT_RIGHT"         , XINE_EVENT_INPUT_RIGHT             },
  { "INPUT_SELECT"        , XINE_EVENT_INPUT_SELECT            },
  { "INPUT_NEXT"          , XINE_EVENT_INPUT_NEXT              },
  { "INPUT_PREVIOUS"      , XINE_EVENT_INPUT_PREVIOUS          },
  { "INPUT_ANGLE_NEXT"    , XINE_EVENT_INPUT_ANGLE_NEXT        },
  { "INPUT_ANGLE_PREVIOUS", XINE_EVENT_INPUT_ANGLE_PREVIOUS    },
  { "INPUT_BUTTON_FORCE"  , XINE_EVENT_INPUT_BUTTON_FORCE      },
  { "INPUT_NUMBER_0"      , XINE_EVENT_INPUT_NUMBER_0          },
  { "INPUT_NUMBER_1"      , XINE_EVENT_INPUT_NUMBER_1          },
  { "INPUT_NUMBER_2"      , XINE_EVENT_INPUT_NUMBER_2          },
  { "INPUT_NUMBER_3"      , XINE_EVENT_INPUT_NUMBER_3          },
  { "INPUT_NUMBER_4"      , XINE_EVENT_INPUT_NUMBER_4          },
  { "INPUT_NUMBER_5"      , XINE_EVENT_INPUT_NUMBER_5          },
  { "INPUT_NUMBER_6"      , XINE_EVENT_INPUT_NUMBER_6          },
  { "INPUT_NUMBER_7"      , XINE_EVENT_INPUT_NUMBER_7          },
  { "INPUT_NUMBER_8"      , XINE_EVENT_INPUT_NUMBER_8          },
  { "INPUT_NUMBER_9"      , XINE_EVENT_INPUT_NUMBER_9          },
  { "INPUT_NUMBER_10_ADD" , XINE_EVENT_INPUT_NUMBER_10_ADD     },
  { "SET_V4L2"            , XINE_EVENT_SET_V4L2                },
  { "PVR_SAVE"            , XINE_EVENT_PVR_SAVE                },
  { "PVR_REPORT_NAME"     , XINE_EVENT_PVR_REPORT_NAME         },
  { "PVR_REALTIME"        , XINE_EVENT_PVR_REALTIME            },
  { "PVR_PAUSE"           , XINE_EVENT_PVR_PAUSE               },
  { "SET_MPEG_DATA"       , XINE_EVENT_SET_MPEG_DATA           },
  /* Following aren't xine's events, just hack */
  { "FOREVER"             , XINE_EVENT_SET_MPEG_DATA + 1       },
  { "NOTHING"             , XINE_EVENT_SET_MPEG_DATA + 1       },
  {  NULL                 , 0                                  }
};

static struct status_struct_s {
  char *name;
  int   status;
} status_struct[] = {
  { "PLAY"  ,  XINE_STATUS_PLAY },
  { "STOP"  ,  XINE_STATUS_STOP },
  { "QUIT"  ,  XINE_STATUS_QUIT },
  { "PAUSE" ,  -1               },
  { "SLOW_4",  -2               },
  { "SLOW_2",  -3               },
  { "NORMAL",  -4               },
  { "FAST_2",  -5               },
  { "FAST_4",  -6               },
  /* Following aren't xine's status */
  { "NONE"  ,  -7               },
  {  NULL   ,  0                }
};

static struct loop_modes_s {
  char *name;
  int   mode;
} loop_modes[] = {
  { "playlist",   PLAYLIST_LOOP     },
  { "mrl"     ,   PLAYLIST_MRL_LOOP },
  { NULL      ,   -1                }
};
  
prompt_state_t prompt_state[] = {
  { XINE_STATUS_IDLE ,  "IDLE"     },
  { XINE_STATUS_STOP ,  "STOP"     },
  { XINE_STATUS_PLAY ,  "PLAY"     },
  { XINE_STATUS_QUIT ,  "QUIT"     },
  { -1               ,  "NOT INIT" },
  { -2               ,  NULL       }
};

static void do_noop(commands_t *, toxine_t *, void *);
static void do_set(commands_t *, toxine_t *, void *);
static void do_get(commands_t *, toxine_t *, void *);
static void do_init(commands_t *, toxine_t *, void *);
static void do_quit(commands_t *, toxine_t *, void *);
static void do_play(commands_t *, toxine_t *, void *);
static void do_stop(commands_t *, toxine_t *, void *);
static void do_pause(commands_t *, toxine_t *, void *);
static void do_jump(commands_t *, toxine_t *, void *);
static void do_seek(commands_t *, toxine_t *, void *);
static void do_shell(commands_t *, toxine_t *, void *);
static void do_version(commands_t *, toxine_t *, void *);
static void do_help(commands_t *, toxine_t *, void *);
static void do_syntax(commands_t *, toxine_t *, void *);
static void do_waitfor(commands_t *, toxine_t *, void *);
static void do_events(commands_t *, toxine_t *, void *);
static void do_playlist(commands_t *, toxine_t *, void *);
static void do_display(commands_t *, toxine_t *, void *);
static void do_mediamk(commands_t *, toxine_t *, void *);
static void do_echo(commands_t *, toxine_t *, void *);
static void do_dumpconfig(commands_t *, toxine_t *, void *);
static void do_dumpstream(commands_t *, toxine_t *, void *);


static commands_t commands[] = {
  { "!",                                   NO_ARGS,       do_shell,
    "Execute a shell.",
    "!"
  },
  { "?",                                   NO_ARGS,       do_help,
    "Display all available commands.",
    "?"
  },
  { "anglenext",                           NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_ANGLE_NEXT event to xine engine",
    "anglenext"
  },
  { "angleprev",                           NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_ANGLE_PREVIOUS event to xine engine",
    "angleprev"
  },
  { "bye",                                 NO_ARGS,       do_quit,
    "Quit toxine",
    "bye"
  },
  { "commands",                            NO_ARGS,       do_help,
    "Display all available commands.",
    "commands"
  },
  { "display",                             REQUIRE_ARGS,  do_display,
    "Display some status",
    "display current <[position | time]>\n"
    "display events <[yes | no | 1 | 0 | true | false] | [status]>\n"
    "display settings"
  },
  { "down",                                NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_DOWN event to xine engine",
    "down"
  },
  { "dumpconfig",                          NO_ARGS,       do_dumpconfig,
    "Dump all entries of current config object.",
    "dumpconfig"
  },
  { "dumpstream",                          NO_ARGS,       do_dumpstream,
    "Dump stream informations (stream info and meta info).",
    "dumpstream"
  },
  { "echo",                                REQUIRE_ARGS,  do_echo,
    "echoing some text / variables",
    "echo <[string|variable]> ...\n"
    "  variable can be:\n"
    "   - $result : last xine call result in int format.\n"
    "   - $cresult: last xine call result in char format.\n"
    "   - $iresult: same as $result.\n"
    "   - $xresult: last xine call result in hexa format.\n"
    "   - $version: toxine's version.\n"
    "   - $date: display date.\n"
    "   - $time: display time.\n"
    "   - $config_file: config filename (*).\n"
    "   - $video_driver: video driver name (*).\n"
    "   - $audio_driver: audio driver name (*).\n"
    "   - $mrl: current mrl (*).\n"
    "   - $script_file: file script name (*).\n"
    " (*): display 'UNSET' when it's unspecified/unavailable.\n"
    "\n"
    "  variable name syntax can be $foo, $(foo) or ${foo}\n"
    ""
  },
  { "exit",                                NO_ARGS,       do_quit,
    "Quit toxine.",
    "exit"
  },
  { "get",                                 OPTIONAL_ARGS, do_get,
    "Get some internal values.",
    "get audio volume\n"
    "get audio mute\n"
    "get audio driver\n"
    "get audio capabilities\n"
    "get video capabilities\n"
    "get video fullscreen\n"
    "get video property <property> (see xine_get_video_property)\n"
    "get video property_min_max <property> (see xine_get_video_property_min_max)\n"
    "get video help"
  },
  { "help",                                OPTIONAL_ARGS, do_help,
    "Display the help text if a command name is specified, otherwise display all available commands.",
    "help [command]"
  },
  { "history",                             NO_ARGS,       do_noop,
    "Display command history.",
    NULL
  },
  { "init",                                NO_ARGS,       do_init,
    "Initialize video/audio/xine-engine.",
    "init"
  },
  { "jump",                                REQUIRE_ARGS,  do_jump,
    "Jump to a new position is steam.",
    "jump <position>"
  },
  { "left",                                NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_LEFT event to xine engine",
    "left"
  },
  { "mediamark",                           REQUIRE_ARGS,  do_mediamk,
    "mediamark management",
    "mediamark show\n"
    "mediamarl load <filename>\n"
    "mediamark save <filename>\n"
    "mediamark loop <[yes | no | 1 | 0 | true | false] | [status]>\n"
    "mediamark delete <[# | ident]> ...\n"
    "mediamark delete <[* | all]>\n"
    "mediamark add [<mrl <mrl>> [[ident <ident>] [start <secs>] [end <secs>]]]\n"
    "mediamark edit <[# | ident]>\n"
    "mediamark select <[# | ident]>"
  },
  { "menu1",                               NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_ANGLE_MENU1 event to xine engine",
    "menu1"
  },
  { "menu2",                               NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_ANGLE_MENU2 event to xine engine",
    "menu2"
  },
  { "menu3",                               NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_ANGLE_MENU3 event to xine engine",
    "menu3"
  },
  { "menu4",                               NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_ANGLE_MENU4 event to xine engine",
    "menu4"
  },
  { "menu5",                               NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_ANGLE_MENU5 event to xine engine",
    "menu5"
  },
  { "menu6",                               NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_ANGLE_MENU6 event to xine engine",
    "menu6"
  },
  { "menu7",                               NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_ANGLE_MENU7 event to xine engine",
    "menu7"
  },
  { "mmk",                                 REQUIRE_ARGS,  do_mediamk,
    "mediamark alias",
    "(see mediamark)"
  },
  { "next",                                NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_NEXT event to xine engine",
    "next"
  },
  { "pause",                               NO_ARGS,       do_pause,
    "Pause/resume playback",
    "pause"
  },
  { "play",                                OPTIONAL_ARGS, do_play,
    "Start playback. mrl can be specified as argument.",
    "play [mrl]"
  },
  { "playlist",                            REQUIRE_ARGS,  do_playlist,
    "playlist management",
    "playlist show\n"
    "playlist select #\n"
    "playlist next\n"
    "playlist previous\n"
    "playlist add <mrl> ...\n"
    "playlist from <plugin>\n"
    "playlist delete # ...\n"
    "playlist delete <[all | *]>\n"
    "playlist load <filename>\n"
    "playlist save <filename>\n"
    "playlist loop <[[yes | no | 1 | 0 | true | false] | [status]]>"
  },
  { "prev",                                NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_PREVIOUS event to xine engine",
    "prev"
  },
  { "quit",                                NO_ARGS,       do_quit,
    "Quit toxine.",
    "quit"
  },
  { "right",                               NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_RIGHT event to xine engine",
    "right"
  },
  { "seek",                                REQUIRE_ARGS,  do_seek,
    "seek to a new position (in seconds) in stream",
    "seek <time>"
  },
  { "select",                              NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_SELECT event to xine engine",
    "select"
  },
  { "set",                                 OPTIONAL_ARGS, do_set,
    "Set some internal values.",
    "set video driver <name>\n"
    "set video fullscreen\n"
    "set video property <property> <value> (see xine_set_video_property)\n"
    "set audio driver <name>\n"
    "set audio volume <level>\n"
    "set audio mute\n"
    "set audio volume <level>\n"
    "set interative <[yes | no | 1 | 0 | true | false]>\n"
    "set mrl <mrl>\n"
    "set loop <[playlist | mrl]>"
  },
  { "shell",                               OPTIONAL_ARGS, do_shell,
    "Execute a shell, or a command if an argument to this command is given.",
    "shell [command]"
  },
  { "stop",                                NO_ARGS,       do_stop,
    "Stop playback",
    "stop"
  },
  { "syntax",                              REQUIRE_ARGS,  do_syntax,
    "Display a command syntax.",
    "syntax <[command]>"
  },
  { "up",                                  NO_ARGS,       do_events,
    "send XINE_EVENT_INPUT_UP event to xine engine",
    "up"
  },
  { "version",                             NO_ARGS,       do_version,
    "Display some version informations",
    "version"
  },
  { "waitfor",                             REQUIRE_ARGS,  do_waitfor,
    "Wait for an event.",
    "waitfor <[UI_PLAYBACK_FINISHED | UI_CHANNELS_CHANGED | UI_SET_TITLE |\n"
    "          UI_MESSAGE | FRAME_FORMAT_CHANGE | AUDIO_LEVEL | QUIT |\n"
    "          PROGRESS | MRL_REFERENCE | UI_NUM_BUTTONS | SPU_BUTTON |\n"
    "          DROPPED_FRAMES | INPUT_MOUSE_BUTTON | INPUT_MOUSE_MOVE |\n"
    "          INPUT_MENU1 | INPUT_MENU2 | INPUT_MENU3 | INPUT_MENU4 |\n"
    "          INPUT_MENU5 | INPUT_MENU6 | INPUT_MENU7 | INPUT_UP |\n"
    "          INPUT_DOWN | INPUT_LEFT | INPUT_RIGHT | INPUT_SELECT |\n"
    "          INPUT_NEXT | INPUT_PREVIOUS | INPUT_ANGLE_NEXT |\n"
    "          INPUT_ANGLE_PREVIOUS | INPUT_BUTTON_FORCE | INPUT_NUMBER_0 |\n"
    "          INPUT_NUMBER_1 | INPUT_NUMBER_2 | INPUT_NUMBER_3 |\n"
    "          INPUT_NUMBER_4 | INPUT_NUMBER_5 | INPUT_NUMBER_6 |\n"
    "          INPUT_NUMBER_7 | INPUT_NUMBER_8 | INPUT_NUMBER_9 |\n"
    "          INPUT_NUMBER_10_ADD | SET_V4L2 | PVR_SAVE |\n"
    "          PVR_REPORT_NAME | PVR_REALTIME | PVR_PAUSE |\n"
    "          SET_MPEG_DATA | FOREVER | NOTHING]>\n"
    "waitfor status <[PLAY | STOP | QUIT | PAUSE | SLOW_4 | SLOW_2 | NORMAL |\n"
    "                FAST_2 | FAST_4]>"
  },
  xine_commands,   /* see xine_commands.[ch] */
  { NULL,         -1,                  NULL, 
    NULL, 
    NULL
  } 
};

/*
 * Callback called on XINE_EVENT_PLAYBACK_FINISHED event
 */
static void finished_cb(toxine_t *tox) {

  if(tox->ignore_finished)
    return;

  tox->playlist.cur++;
  
 __playloop:
  if (tox->playlist.cur < tox->playlist.num) {
    
    if(tox->playlist.loop && (tox->loop_mode == PLAYLIST_MRL_LOOP))
      tox->playlist.cur--;

  __playmrloop:
    toxine_set_current_mrl_from_cur(tox);
    if(xine_open(tox->stream, tox->current_mrl)) {
      tox->xine_state |= XINE_OPEN;
      _xine_play(NULL, tox, (void *)NO_ARGS);
    }
  }
  else {
    if(tox->playlist.loop && tox->playlist.num) {
      
      if(tox->loop_mode == PLAYLIST_MRL_LOOP) {
	tox->playlist.cur--;
	goto __playmrloop;
      }
      else if(tox->loop_mode == PLAYLIST_LOOP) {
	tox->playlist.cur = 0;
      }

      goto __playloop;
    }
    else
      tox->playlist.cur--;
  }

}

void *toxine_end_thread(void *data) {
  toxine_t *tox = (toxine_t *) data;
  int       cur;

  tox->playlist.thread_num++;
  cur = tox->playlist.cur;

  while(tox->running && (tox->event.last.type != XINE_EVENT_UI_PLAYBACK_FINISHED)) {
    
    xine_usec_sleep(500000.0);
 
    if((tox->xine_state & (XINE_NEW | XINE_OPEN))) {
      int secs;

      secs = xine_get_pos_length(tox->stream, NULL, &secs, NULL);
      secs /= 1000;

      if(cur == tox->playlist.cur) {
	
	if(secs >= tox->playlist.mmk[cur]->end) {
	  tox->ignore_finished = 0;
	  finished_cb(tox);
	  goto __end_done;
	}

      }
      else
	goto __end_done;

    }
    else
      goto __end_done;

  }
  
 __end_done:
  tox->playlist.thread_num--;
  pthread_exit(NULL);
}

/*
 * Listener of xine engine events.
 */
void toxine_event_listener(void *data, const xine_event_t *event) {
  toxine_t *tox = (toxine_t *) data;
  int       i;
  char     *event_name = NULL;

  pthread_mutex_lock(&tox->event.mutex);
  tox->event.last.type = event->type;
  pthread_mutex_unlock(&tox->event.mutex);
  
  switch(tox->event.last.type) { 
  case XINE_EVENT_UI_PLAYBACK_FINISHED:
    finished_cb(tox);
    tox->ignore_finished = 0;
    break;
  default:
    break;
  }

  if(tox->display_events) {
    pinfo("Get xine event:\n");
    poutalign();
    for(i = 0; event_struct[i].name != NULL; i++) {
      if(event_struct[i].event == tox->event.last.type) {
	event_name = event_struct[i].name;
	break;
      }
    }
    pout("%s", (event_name) ? event_name : "UNKNOWN EVENT");
    pinfo("\n");
  }
 
  if(tox->video.cur_plugin && tox->video.cur_plugin->receive_xine_event)
    tox->video.cur_plugin->receive_xine_event(tox, event);
  
}

/*
 * Initialize video out plugin.
 */
int toxine_video_out_init(toxine_t *tox) {
  int                  i = 0;
  pthread_attr_t       pth_attrs;
  struct sched_param   pth_params;
  
  if(!tox->video.name) {
    perr("you should specify a video driver name first.\n");
    return 0;
  }
  
  if(!tox->video.plugins_num) {
    perr("no video plugin found.\n");
    return 0;
  }
  
  if(tox->video.cur_plugin) {
    perr("video plugin already initialized.\n");
    return 1;
  }
  else {
    for(i = 0; i < tox->video.plugins_num; i++) {
      char **vonames = (char **)tox->video.plugins[i]->get_names();
      int j;
      
      for(j = 0; vonames[j] != NULL; j++) {
	if(!strcasecmp(vonames[j], tox->video.name)) {
	  tox->video.cur_plugin = tox->video.plugins[i];
	  goto __vo_found;
	}
      }
    }
    
  __vo_found:
    
    if(!tox->video.cur_plugin) {
      perr("No video driver name match with '%s'\n", tox->video.name);
      return 0;
    }

    if(tox->video.last_plugin && (tox->video.last_plugin != tox->video.cur_plugin) && tox->video.last_plugin->video_out_release)
      tox->video.last_plugin->video_out_release(tox);
    
    tox->video.last_plugin = tox->video.cur_plugin;
    
    tox->video.running = 1;
    tox->video.cur_plugin->video_out_init(tox);

    /* Display plugin help string, if available */
    if(tox->video.cur_plugin->get_help) {
      const char *help = tox->video.cur_plugin->get_help();
      
      if(help) {
	pinfo("Video out plugin help:\n");
	pinfo("%s\n", help);
      }
    }
    
    tox->video.loop_thread = 0;
    if(tox->video.cur_plugin->video_out_event_loop) {
      pthread_attr_init(&pth_attrs);
      pthread_attr_getschedparam(&pth_attrs, &pth_params);
      pth_params.sched_priority = sched_get_priority_min(SCHED_OTHER);
      pthread_attr_setschedparam(&pth_attrs, &pth_params);
      
      pthread_create(&tox->video.loop_thread, &pth_attrs, 
		     tox->video.cur_plugin->video_out_event_loop, (void *)tox);
    }
      
  }
  
  return 1;
}

/*
 * Initialize audio plugin.
 */
int toxine_audio_out_init(toxine_t *tox) {

  if(!tox->audio.name) {
    perr("no audio driver sets, no audio output.\n");
    return 1;
  }
  
  if(tox->audio.port == NULL) {

    tox->audio.mixer.enable = 0;
    tox->audio.mixer.caps   = 0;
    
    tox->audio.port = xine_open_audio_driver(tox->xine, (const char*)tox->audio.name, NULL);
    if(tox->audio.port == NULL) {
      perr("xine_load_audio_output_plugin() failed to load '%s' driver.\n", tox->audio.name);
      return 0;
    }
    tox->audio.mixer.enable = 1;
    tox->audio.mixer.caps   = 1;
  }
  
  return 1;
}

/*
 * Display help of given command.
 */
void toxine_command_help(commands_t *command, toxine_t *tox) {

  if(command) {
    if(command->help)
      phelp("Help of '%s' command:\n%s\n", command->command, command->help);
    else
      pinfo("There is no help text for command '%s'\n", command->command);
  }
}

/*
 * Display syntax of given command.
 */
void toxine_command_syntax(commands_t *command, toxine_t *tox) {
  
  if(command) {
    if(command->syntax)
      phelp("Syntax of '%s' command:\n%s\n", command->command, command->syntax);
    else
      pinfo("There is no syntax definition for command '%s'\n", command->command);
  }
}

/*
 * Handle signals event.
 */
static void toxine_signals_handler (int sig) {

  switch(sig) {
    /* Kill the line on C-c */
  case SIGINT:
  case SIGTERM:
    if(rl_prompt) { /* readline is running, otherwise we are in script mode */
      rl_kill_full_line(1, 1);
      rl_redisplay();
    }
    else
      exit(1);
    break;

  }
}

/*
 * Handle user entered commands.
 */
void toxine_handle_command(toxine_t *tox, void *data) {
  int i, found;

  /* pass command line to the chainsaw */

  do {

    toxine_parse_command(tox);
    
    if(strlen(tox->command.line)) {
      
      /*
       * Is the command match with one available ?
       */
      i = found = 0;
      while((commands[i].command != NULL) && (found == 0)) {
	if(!strncasecmp(tox->command.command, commands[i].command, strlen(tox->command.command))) {
	  if((commands[i].argtype == REQUIRE_ARGS) 
	     && (tox->command.num_args <= 0)) {
	    perr("Command '%s' require argument(s).\n", commands[i].command);
	    found++;
	  }
	  else if((commands[i].argtype == NO_ARGS) && (tox->command.num_args > 0)) {
	    perr("Command '%s' doesn't require argument.\n", commands[i].command);
	    found++;
	  }
	  else {
	    commands[i].function(&commands[i], tox, NULL);
	    found++;
	  }
	}
	i++;
      }
      
      if(!found)
	perr("unhandled command '%s'.\n", tox->command.command);
      
    }
    
    toxine_free(tox->command.line);

  } while((tox->command.remain != NULL));
}

/*
 * completion generator functions.
 */
static char *toxine_command_generator(const char *text, int state) {
  static int   index, len;
  char        *cmd, *retcmd = NULL;
  
  if(!state) {
    index = 0;
    len = strlen(text);
  }

  if(len) {
    while((cmd = commands[index].command) != NULL) {
      index++;
      if(strncasecmp(cmd, text, len) == 0) {
	retcmd = strdup(cmd);
	return retcmd;
      }
    }
  }
  
  return NULL;
}
static char **toxine_completion_function(const char *text, int start, int end) {
  char  **cmd = NULL;
  
  if(start == 0)
    cmd = rl_completion_matches (text, toxine_command_generator);
  
  return cmd;
}

/*
 * Called when we leave toxine.
 */
void toxine_stop(toxine_t *tox) {

  if(tox->xine) {
    toxine_set_command_line(tox, "xine_stop");
    toxine_handle_command(tox, NULL);
  }
}

/*
 * The main loop (except for execute and script mode).
 */
void toxine_run(toxine_t *tox, int argc, char **argv) {
  char              *grabbed_line;
  struct sigaction   action;

  /*
   * install sighandler on SIGINT signal, i don't want to 
   * quit toxine that way.
   */
  action.sa_handler = toxine_signals_handler;
  sigemptyset(&(action.sa_mask));
  action.sa_flags = 0;
  if(sigaction(SIGHUP, &action, NULL) != 0)
    fprintf(stderr, "sigaction(SIGHUP) failed: %s\n", strerror(errno));

  action.sa_handler = toxine_signals_handler;
  sigemptyset(&(action.sa_mask));
  action.sa_flags = 0;
  if(sigaction(SIGUSR1, &action, NULL) != 0)
    fprintf(stderr, "sigaction(SIGUSR1) failed: %s\n", strerror(errno));

  action.sa_handler = toxine_signals_handler;
  sigemptyset(&(action.sa_mask));
  action.sa_flags = 0;
  if(sigaction(SIGUSR2, &action, NULL) != 0)
    fprintf(stderr, "sigaction(SIGUSR2) failed: %s\n", strerror(errno));

  action.sa_handler = toxine_signals_handler;
  sigemptyset(&(action.sa_mask));
  action.sa_flags = 0;
  if(sigaction(SIGINT, &action, NULL) != 0)
    fprintf(stderr, "sigaction(SIGINT) failed: %s\n", strerror(errno));

  action.sa_handler = toxine_signals_handler;
  sigemptyset(&(action.sa_mask));
  action.sa_flags = 0;
  if(sigaction(SIGTERM, &action, NULL) != 0)
    fprintf(stderr, "sigaction(SIGTERM) failed: %s\n", strerror(errno));

  action.sa_handler = toxine_signals_handler;
  sigemptyset(&(action.sa_mask));
  action.sa_flags = 0;
  if(sigaction(SIGQUIT, &action, NULL) != 0)
    fprintf(stderr, "sigaction(SIGQUIT) failed: %s\n", strerror(errno));

  action.sa_handler = toxine_signals_handler;
  sigemptyset(&(action.sa_mask));
  action.sa_flags = 0;
  if(sigaction(SIGALRM, &action, NULL) != 0)
    fprintf(stderr, "sigaction(SIGALRM) failed: %s\n", strerror(errno));

  if(argc) {
    int i;

    tox->playlist.cur = 0;
    
    for(i = 0; i < argc; i++)
      playlist_add_entry(tox, argv[i], argv[i], 0, -1);
  }

  toxine_load_vo_plugins(tox);

  if(tox->autoinit) {
    _xine_init(NULL, tox, NULL);
    if(tox->xine)
      _xine_stream_new(NULL, tox, NULL);
  }
  
  if(tox->script.in_use) {
    toxine_handle_script(tox);
    tox->video.running = tox->running = 0;
  }
  else {
    
    if(tox->command.execute) {
      int old_interactive = tox->interactive;

      tox->interactive = -1;
      toxine_handle_command(tox, NULL);
      /* is script changed interactive ?? */
      if(tox->interactive == -1)
	tox->interactive = old_interactive;
      
    }
    
    grabbed_line = NULL;
    toxine_update_prompt(tox);
    
    /* Few realdline inits */
    rl_readline_name = "toxine";
    rl_initialize();
    rl_attempted_completion_function = (CPPFunction *)toxine_completion_function;

    while(tox->running) {
      
    __re_readline:
      if((grabbed_line = readline(tox->prompt)) == NULL) {
	fprintf(stderr, "%s(%d): readline() failed: %s\n",
		__tox_func__, __LINE__, strerror(errno));
	goto __re_readline;
      }

      if (*grabbed_line) {
	char *line;
	
	line = grabbed_line;
	
	toxine_set_command_line(tox, line);
	toxine_handle_command(tox, NULL);
	
	add_history(toxine_atoa(line));
	
	/* Display command history */
	if(!strncasecmp(line, "history", strlen(line))) {
	  HIST_ENTRY **list = history_list ();
	  int i;
	  
	  if(list) {
	    for(i = 0; list[i]; i++)
	      pinfo("%d: %s\r\n", i, list[i]->line);
	  }
	}
	
	toxine_free(grabbed_line);
	
      }
      toxine_update_prompt(tox);
    }

  }

  /* deinit video_out plugin */
  if(tox->xine_state & XINE_INIT) {
    
    if(tox->playlist.thread_num > 0)
      pthread_join(tox->playlist.end_thread, NULL);

    _xine_exit(NULL, tox, NULL);
  }

  if(tox->video.last_plugin && tox->video.last_plugin->video_out_release)
    tox->video.last_plugin->video_out_release(tox);
  
}

/*
 * **********************************************************************************************
 * ******                                   COMMANDS                                      *******
 * **********************************************************************************************
 */

/*
 * get values.
 */
static void do_get(commands_t *command, toxine_t *tox, void *data) {
  int nargs;

  nargs = toxine_is_args(tox);
  if(nargs) {
    if(toxine_is_arg_contain(tox, 1, "audio")) {
      if(nargs >= 2) {
	
	if(toxine_is_arg_contain(tox, 2, "volume")) {
	  
	  if(tox->audio.mixer.enable)
	    toxine_draw_bar("Current audio volume", 0, 100, tox->audio.mixer.volume_level);
	  else
	    pinfo("Audio mixer is disable.\n");
	  
	}
	else if(toxine_is_arg_contain(tox, 2, "mute")) {
	  
	  if(tox->audio.mixer.enable)
	    pinfo("Audio mute status: %s\n", (tox->audio.mixer.mute)?"muted":"unmuted");
	  else
	    pinfo("Audio mixer is disable.\n");

	}
	else if(toxine_is_arg_contain(tox, 2, "driver") || toxine_is_arg_contain(tox, 2, "name"))
	  pinfo("current audio driver name: %s\n", (tox->audio.name)?tox->audio.name:"not set");

      }
    }
    else if(toxine_is_arg_contain(tox, 1, "video")) {

      if(nargs == 2) {

	if(toxine_is_arg_contain(tox, 2, "fullscreen")) {
	  int fullscreen;
	  
	  return_if_no_init(tox);

	  if(tox->video.cur_plugin->is_fullscreen) {
	    fullscreen = tox->video.cur_plugin->is_fullscreen(tox);
	    pinfo("fullscreen is %d\n", fullscreen);
	    pinfo(".\n");
	  }
	  else
	    pout("fullscreen isn't supported by video plugin\n");
	}
	else if(toxine_is_arg_contain(tox, 2, "help")) {

	  if(tox->video.cur_plugin && tox->video.cur_plugin->get_help) {
	    const char *help = tox->video.cur_plugin->get_help();
	    
	    if(help) {
	      pinfo("Video out plugin help:\n");
	      pinfo("%s\n", help);
	    }
	  }
    
	}


      }
      else if(nargs >= 3) {
	
	if(toxine_is_arg_contain(tox, 2, "property")) {
	  char buffer[1024];
	  
	  memset(&buffer, 0, sizeof(buffer));
	  sprintf(buffer, "%s %s", "xine_get_video_property", toxine_get_arg(tox, 3));
	  
	  toxine_set_command_line(tox, buffer);
	  toxine_handle_command(tox, NULL);

	}
	else if(toxine_is_arg_contain(tox, 2, "property_min_max")) {
	  char buffer[1024];
	  
	  memset(&buffer, 0, sizeof(buffer));
	  sprintf(buffer, "%s %s", "xine_get_video_property_min_max", toxine_get_arg(tox, 3));

	  toxine_set_command_line(tox, buffer);
	  toxine_handle_command(tox, NULL);

	}

      }
    }
  }
}

/*
 * set values
 */
static void do_set(commands_t *command, toxine_t *tox, void *data) {
  int nargs;

  nargs = toxine_is_args(tox);
  if(nargs) {
    if(nargs >= 2) {

      if(toxine_is_arg_contain(tox, 1, "mrl"))
	toxine_set_current_mrl(tox, (toxine_get_arg(tox, 2)));
      if(toxine_is_arg_contain(tox, 1, "loop")) {
	const char *lmode = toxine_get_arg(tox, 2);
	int         i, mode = -1;
	
	for(i = 0; loop_modes[i].name != NULL; i++) {
	  if(!strncasecmp(lmode, loop_modes[i].name, strlen(lmode))) {
	    mode = loop_modes[i].mode;
	    break;
	  }
	}

	if(mode >= 0) {
	  tox->loop_mode = mode;
	  pinfo("Loop mode is set to: '%s' [%d]\n", loop_modes[i].name, loop_modes[i].mode); 
	}
	else
	  perr("Loop mode '%s' isn't valid\n", lmode);

      }
      else if(toxine_is_arg_contain(tox, 1, "interactive")) {
	tox->interactive = toxine_get_bool_value((toxine_get_arg(tox, 2)));
      }
      else if(toxine_is_arg_contain(tox, 1, "video")) {

	if(nargs >= 2) {
	  if(toxine_is_arg_contain(tox, 2, "property")) {

	    if(nargs >= 4) {
	      char buffer[1024];

	      memset(&buffer, 0, sizeof(buffer));
	      sprintf(buffer, "%s %s %s",
		      "xine_set_param", toxine_get_arg(tox, 3), toxine_get_arg(tox, 4));

	      toxine_set_command_line(tox, buffer);
	      toxine_handle_command(tox, NULL);

	    }

	  }
	  else if(toxine_is_arg_contain(tox, 2, "driver")) {

	    if(nargs >= 3) {
	      if(tox->xine_state & XINE_INIT) {
		perr("xine engine already initialized.\n");
		return;
	      }
	      
	      if(tox->video.name) 
		free((char *)tox->video.name);

	      tox->video.name = strdup(toxine_get_arg(tox, 3));

	      pinfo("new selected video driver: %s\n", tox->video.name);
	    }
	    else
	      pinfo("current video driver name: %s\n", 
		    (tox->video.name) ? tox->video.name : "unset");
	    
	  }
	  else if(toxine_is_arg_contain(tox, 2, "fullscreen")) {
	    return_if_no_init(tox);
	    if(tox->video.cur_plugin->fullscreen)
	      tox->video.cur_plugin->fullscreen(tox);
	  }
	}
      }
      else if(toxine_is_arg_contain(tox, 1, "audio")) {
	
	if(nargs >= 2) {
	  if(toxine_is_arg_contain(tox, 2, "driver")) {

	    if(nargs >= 3) {
	      if(tox->xine_state & XINE_INIT) {
		perr("xine engine already initialized.\n");
		return;
	      }

	      if(tox->audio.name) 
		free((char *)tox->audio.name);

	      tox->audio.name = strdup(toxine_get_arg(tox, 3));
	      pinfo("new selected audio driver: %s\n", tox->audio.name);
	    }
	    else
	      pinfo("current video driver name: %s\n", 
		    (tox->audio.name) ? tox->audio.name : "unset");

	  }
	  else if(toxine_is_arg_contain(tox, 2, "volume")) {
	    if(nargs >= 3) {
	      /* SET AUDIO VOLUME */
	      if(tox->audio.mixer.enable) {
		char buffer[1024];
		
		tox->audio.mixer.volume_level = atoi(toxine_get_arg(tox, 3));
		
		sprintf(buffer, "%s %d", 
			"xine_set_param XINE_PARAM_AUDIO_VOLUME", tox->audio.mixer.volume_level);

		toxine_set_command_line(tox, buffer);
		toxine_handle_command(tox, NULL);

	      }
	    }      
	  }
	  else if(toxine_is_arg_contain(tox, 2, "mute")) {

	    toxine_set_command_line(tox, "xine_set_param XINE_PARAM_AUDIO_MUTE");
	    toxine_handle_command(tox, NULL);

	  }
	}
      }
    }
  }

}

/*
 * dummy function.
 */
static void do_noop(commands_t *command, toxine_t *tox, void *data) {
}

/*
 * initialize xine engine
 */
static void do_init(commands_t *command, toxine_t *tox, void *data) {

  _xine_init(command, tox, NULL);

  if(tox->xine)
    _xine_stream_new(command, tox, NULL);
}

/*
 * leave toxine.
 */
static void do_quit(commands_t *command, toxine_t *tox, void *data) {
  
  if(tox->interactive) {
    if((toxine_confirm("Do you really want to quit toxine ? [yes|no]: ")))
      tox->video.running = tox->running = 0;
  }
  else
    tox->video.running = tox->running = 0;

}

/*
 * Start to playback an mrl.
 */
static void do_play(commands_t *command, toxine_t *tox, void *data) {

  return_if_no_init(tox);
  return_if_no_stream(tox);

  if(tox->command.num_args)
    toxine_set_current_mrl(tox, (toxine_get_arg(tox, 1)));
  else if(!tox->current_mrl)
    toxine_set_current_mrl_from_cur(tox);
  
  if(!tox->current_mrl) {
    perr("No mrl set.\n");
    return;
  }
  
  if(xine_open(tox->stream, tox->current_mrl)) {
    tox->xine_state |= XINE_OPEN;
    _xine_play(command, tox, (void *)NO_ARGS);
  }

}

/*
 * Stop playback.
 */
static void do_stop(commands_t *command, toxine_t *tox, void *data) {
  
  return_if_no_init(tox);

  _xine_stop(command, tox, NULL);
}

/*
 * Pause/Resume playback.
 */
static void do_pause(commands_t *command, toxine_t *tox, void *data) {

  return_if_no_init(tox);

  if((xine_get_param(tox->stream, XINE_PARAM_SPEED)) != XINE_SPEED_PAUSE)
    xine_set_param(tox->stream, XINE_PARAM_SPEED, XINE_SPEED_PAUSE);
  else
    xine_set_param(tox->stream, XINE_PARAM_SPEED, XINE_SPEED_NORMAL);
  
}

/*
 * Jump in stream.
 */
static void do_jump(commands_t *command, toxine_t *tox, void *data) {
  int pos;
  
  return_if_no_init(tox);

  if((xine_get_stream_info(tox->stream, XINE_STREAM_INFO_SEEKABLE)) == 0)
    return;
  
  if(!tox->current_mrl) {
    perr("You should set/specify an MRL.\n");
    return;
  }
  
  pos = atoi((toxine_get_arg(tox, 1)));
  
  if(!xine_play(tox->stream, pos, 0))
    fprintf(stderr, "xine_play() failed.\n");
  
}

/*
 * Seek in stream.
 */
static void do_seek(commands_t *command, toxine_t *tox, void *data) {
  int msec;

  return_if_no_init(tox);
  
  if((xine_get_stream_info(tox->stream, XINE_STREAM_INFO_SEEKABLE)) == 0)
    return;
  
  if(!tox->current_mrl) {
    perr("You should set/specify an MRL.\n");
    return;
  }
  
  msec = (atoi((toxine_get_arg(tox, 1)))) * 1000;

  if(msec < 0)
    msec = 0;
  
  if(!xine_play(tox->stream, 0, msec))
    fprintf(stderr, "xine_play() failed.\n");
}

/*
 * Execute a shell command (or a shell itself)
 */
static void do_shell(commands_t *command, toxine_t *tox, void *data) {
  char buffer[2048] = "";
  int  retval;
  
  if(tox->command.num_args) {
    int i;
    
    for(i = 0; i < tox->command.num_args; i++)
      sprintf(buffer, "%s %s", buffer, (toxine_get_arg(tox, (i+1))));
  }
  else
    sprintf(buffer, "%s", "/bin/sh --login");
  
  retval = toxine_system(0, buffer);
  pinfo("shell returned %d.\n", retval);
  pinfo(".\n");
}

/*
 * Display some version informations.
 */
static void do_version(commands_t *command, toxine_t *tox, void *data) {
  int major, minor, sub;

  pinfo("Toxine version %s\n", VERSION);
  pinfo("Copyright 2002-2004, Daniel Caujolle-Bert <f1rmb@users.sourceforge.net>.\n");
  pinfo("Built with xine library %d.%d.%d (%s).\n",
	XINE_MAJOR_VERSION, XINE_MINOR_VERSION, XINE_SUB_VERSION, XINE_VERSION);
  xine_get_version(&major, &minor, &sub);
  pinfo("Found xine library version: %d.%d.%d (%s).\n", 
	major, minor, sub, xine_get_version_string());
  pinfo("Using readline version %s.\n", rl_library_version);
  pinfo(".\n");
}

/*
 * Display syntaxe of a given command.
 */ 
static void do_syntax(commands_t *command, toxine_t *tox, void *data) {
  int i;
  
  for(i = 0; commands[i].command != NULL; i++) {
    if(!strcasecmp((toxine_get_arg(tox, 1)), commands[i].command)) {
      toxine_command_syntax(&commands[i], tox);
      return;
    }
  }

  perr("Unknown command '%s'\n", (toxine_get_arg(tox, 1)));
}

/*
 * Display help of a given command.
 */
static void do_help(commands_t *command, toxine_t *tox, void *data) {

  if(!tox->command.num_args) {
    int i = 0, j;
    int maxlen = 0;
    int curpos = 0;
    int screen_rows, screen_cols;

    rl_get_screen_size(&screen_rows, &screen_cols);

    while(commands[i].command != NULL) {
      if(strlen(commands[i].command) > maxlen)
	maxlen = strlen(commands[i].command);
      i++;
    }
    
    maxlen++; 
    i = 0;  
    
    pinfo("Available commands are:\n");
    poutalign();
    curpos += strlen(EMPTY_PREFIX);
    
    while(commands[i].command != NULL) {
      
      if((curpos + maxlen) >= screen_cols) {
	pout("\n");
	poutalign();
	curpos = strlen(EMPTY_PREFIX);
      }
      
      pout("%s", commands[i].command);
      curpos += strlen(commands[i].command);
      
      for(j = 0; j < (maxlen - strlen(commands[i].command)); j++) {
	pout(" ");
	curpos++;
      }
      
      i++;
    }
    
    pout("\n");
    pinfo(".\n");
  }
  else {
    int i;
    
    for(i = 0; commands[i].command != NULL; i++) {
      if(!strcasecmp((toxine_get_arg(tox, 1)), commands[i].command)) {
	toxine_command_help(&commands[i], tox);
	return;
      }
    }

    perr("Unknown command '%s'\n", (toxine_get_arg(tox, 1)));
  }
}

/*
 * boolean, return 1 of name match with a valid event.
 */
static int toxine_get_event_value(const char *name, uint32_t *value) {
  int i;

  if((!name) || (!value))
    return 0;

  for(i = 0; event_struct[i].name != NULL; i++) {
    if(!strncasecmp(name, event_struct[i].name, strlen(name))) {
      *value = event_struct[i].event;
      return 1;
    }
  }

  return 0;
}

/*
 * boolean, return 1 is status is valid.
 */
static int toxine_is_status_valid(const char *status, int *value) {
  int i;
  
  if((!status) || (!value))
    return 0;
  
  for(i = 0; status_struct[i].name != NULL; i++) {
    if(!strncasecmp(status, status_struct[i].name, strlen(status))) {
      *value = status_struct[i].status;
      return 1;
    }
  }

  return 0;
}

/*
 * wait for something.
 */
static void do_waitfor(commands_t *command, toxine_t *tox, void *data) {
  uint32_t event_value;
  int      event_happend = 0;
  int      status, status_value, status_happend = 0;
  int      speed, speed_status;

  return_if_no_init(tox);

  if(toxine_get_event_value((toxine_get_arg(tox, 1)), &event_value)) {
    
    do {
      xine_usec_sleep((1/90000.0)*2);
      pthread_mutex_lock(&tox->event.mutex);
      event_happend = (tox->event.last.type == event_value) ? 1 : 0;
      pthread_mutex_unlock(&tox->event.mutex);
    } while(!event_happend);

    pthread_mutex_lock(&tox->event.mutex);
    tox->event.last.type = 0;
    pthread_mutex_unlock(&tox->event.mutex);
  }
  else if((toxine_is_arg_contain(tox, 1, "status")) && (toxine_is_args(tox) >= 2)) {
    if(toxine_is_status_valid((toxine_get_arg(tox, 2)), &status_value)) {
      if(status_value >= 0) { /* STOP PLAY QUIT */
	do {
	  xine_usec_sleep((1/90000.0)*2);
	  status = xine_get_status(tox->stream);
	  status_happend = (status == status_value) ? 1 : 0;
	} while(!status_happend);
	
      }
      else { /* SPEED */

	/* Get expected speed */
	switch(status_value) {
	case -1:
	  speed = XINE_SPEED_PAUSE;
	  break;
	case -2:
	  speed = XINE_SPEED_SLOW_4;
	  break;
	case -3:
	  speed = XINE_SPEED_SLOW_2;
	  break;
	case -4:
	  speed = XINE_SPEED_NORMAL;
	  break;
	case -5:
	  speed = XINE_SPEED_FAST_2;
	  break;
	case -6:
	  speed = XINE_SPEED_FAST_4;
	  break;
	default:
	  perr("Unhandled speed status.\n");
	  return;
	  break;
	}
	
	do {
	  
	  xine_usec_sleep((1/90000.0)*2);
	  speed_status = xine_get_param(tox->stream, XINE_PARAM_SPEED);
	  status_happend = (speed == speed_status) ? 1 : 0;
	  
	} while(!status_happend);
	
      }
    }
    else
      perr("Invalid status name '%s'\n", (toxine_get_arg(tox, 2)));
  }
  else
    perr("Wait for event '%s' is invalid.\n", (toxine_get_arg(tox, 1)));
}

static void do_events(commands_t *command, toxine_t *tox, void *data) {
  int           i;
  static struct event_aliases_s {
    char   *alias;
    int     event;
  } event_aliases[] = {
    { "menu1",     XINE_EVENT_INPUT_MENU1          },
    { "menu2",     XINE_EVENT_INPUT_MENU2          },
    { "menu3",     XINE_EVENT_INPUT_MENU3          },
    { "menu4",     XINE_EVENT_INPUT_MENU4          },
    { "menu5",     XINE_EVENT_INPUT_MENU5          },
    { "menu6",     XINE_EVENT_INPUT_MENU6          },
    { "menu7",     XINE_EVENT_INPUT_MENU7          },
    { "up",        XINE_EVENT_INPUT_UP             },
    { "down",      XINE_EVENT_INPUT_DOWN           },
    { "left",      XINE_EVENT_INPUT_LEFT           },
    { "right",     XINE_EVENT_INPUT_RIGHT          },
    { "next",      XINE_EVENT_INPUT_NEXT           },
    { "prev",      XINE_EVENT_INPUT_PREVIOUS       },
    { "anglenext", XINE_EVENT_INPUT_ANGLE_NEXT     },
    { "angleprev", XINE_EVENT_INPUT_ANGLE_PREVIOUS },
    { "select",    XINE_EVENT_INPUT_SELECT         },
    { NULL,        0                               }
  };
  
  for(i = 0; event_aliases[i].alias != NULL; i++) {
    if(!strncmp(command->command, event_aliases[i].alias, strlen(command->command))) {
      _xine_event_send(command, tox, (void *) event_aliases[i].event);
      break;
    }
  }
}

static void do_playlist(commands_t *command, toxine_t *tox, void *data) {
  int nargs;

  nargs = toxine_is_args(tox);

  if(nargs == 1) {
    if(toxine_is_arg_contain(tox, 1, "show")) {

      if(tox->playlist.num) {
	int i;
	
	pinfo("Playlist (* is current):\n");
	for(i = 0; i < tox->playlist.num; i++) {

	  poutalign();

	  if(tox->playlist.cur == i)
	    pout("*%4d: '%s' ", i, tox->playlist.mmk[i]->mrl);
	  else
	    pout(" %4d: '%s' ", i, tox->playlist.mmk[i]->mrl);

	  pout("[%d .. %d]\n", tox->playlist.mmk[i]->start, tox->playlist.mmk[i]->end);

	}

	pinfo(".\n");

      }
      else
	pinfo("Empty playlist.\n");
    }
    else if(toxine_is_arg_contain(tox, 1, "next")) {
      tox->ignore_finished = 0;
      finished_cb(tox);
    }
    else if(toxine_is_arg_contain(tox, 1, "prev")) {

      tox->ignore_finished = 1;
      tox->playlist.cur--;

      if((tox->playlist.cur >= 0) && (tox->playlist.cur < tox->playlist.num)) {
	
	toxine_set_current_mrl_from_cur(tox);
	_xine_play(NULL, tox, (void *)NO_ARGS);
	
      }
      else {
	tox->playlist.cur = 0;
	tox->ignore_finished = 0;
      }
      
    }
    
  }
  else if(nargs >= 2) {

    if(toxine_is_arg_contain(tox, 1, "select")) {
      int entry = atoi(toxine_get_arg(tox, 2));
      int old_cur = tox->playlist.cur;
      
      if(entry < tox->playlist.num) {
	tox->playlist.cur = entry;
	toxine_set_current_mrl_from_cur(tox);
      }

      /*
       * If xine engine is in PLAY status, and new selected entry is != than old one, 
       * start playback with new selected stream 
       */
      if(tox->xine && (old_cur != tox->playlist.cur)) {
	int status = xine_get_status(tox->stream);
	
	if(status == XINE_STATUS_PLAY)
	  _xine_play(NULL, tox, (void *)NO_ARGS);
	
      }

    }
    else if(toxine_is_arg_contain(tox, 1, "add")) {
      const char *mrl;
      int         i = 2;

      while((mrl = toxine_get_arg(tox, i)) != NULL) {
	
	//if(strlen(toxine_atoa(mrl))) {
	if(strlen(mrl)) {
	  char *expanded_mrl;
	  
	  expanded_mrl = tilde_expand(mrl);
	  
	  playlist_add_entry(tox, expanded_mrl, NULL, 0, -1);
	  
	  free(expanded_mrl);
	  
	  if(tox->playlist.cur == -1) {
	    tox->playlist.cur = 0;
	    toxine_set_current_mrl_from_cur(tox);
	  }
	}
	i++;
      }
    }
    else if(toxine_is_arg_contain(tox, 1, "from")) {
      char             **ap;
      int                i, mrls;
      const char        *plugname;
      
      return_if_no_init(tox);
      
      plugname = toxine_get_arg(tox, 2);
      pinfo("Grab mrls from '%s' input plugin.\n", plugname);
      ap = xine_get_autoplay_mrls(tox->xine, plugname, &mrls);
      if(ap && mrls) {

	for(i = 0; i < mrls; i++)
	  playlist_add_entry(tox, ap[i], NULL, 0, -1);

	if(tox->playlist.num && (tox->playlist.cur == -1)) {
	  tox->playlist.cur = 0;
	  toxine_set_current_mrl_from_cur(tox);
	}

	pinfo("Plugin '%s' returns %d mrls\n", plugname, tox->playlist.num);
      }
      else
	pinfo("Plugin '%s' returned nothing.\n", toxine_get_arg(tox, 2));
      
      pinfo(".\n");
    }
    else if(toxine_is_arg_contain(tox, 1, "delete")) {
      
      if(toxine_is_arg_contain(tox, 2, "all") || toxine_is_arg_contain(tox, 2, "*")) {
	if(tox->playlist.num) {
	  
	  for(; tox->playlist.num > 0;)
	    playlist_free_entry(tox, (tox->playlist.num - 1));

	  toxine_free(tox->playlist.mmk);

	  tox->playlist.num = 0;
	  tox->playlist.cur = -1;
	  tox->playlist.loop = 0;

	  toxine_set_current_mrl_from_cur(tox);
	}
      }
      else {
	const char *argument;
	int         i = 2;
	
	while((argument = toxine_get_arg(tox, i))) {
	  int entry = atoi(argument);
	  
	  if(tox->playlist.num && (entry < tox->playlist.num)) {
	    int j;
	    
	    if(tox->playlist.cur >= entry) {
	      tox->playlist.cur = -1;
	      tox->current_mrl = NULL;
	    }
	    
	    playlist_free_entry(tox, entry);
	    
	    for(j = entry; j < tox->playlist.num; j++)
	      tox->playlist.mmk[j] = tox->playlist.mmk[j + 1];

	    tox->playlist.mmk = (mediamark_t **) 
	      realloc(tox->playlist.mmk, sizeof(mediamark_t *) * (tox->playlist.num + 2));
	    
	    tox->playlist.mmk[tox->playlist.num] = NULL;
	    
	    if(!tox->playlist.num)
	      tox->playlist.loop = 0;
	    
	    if(tox->playlist.num && (tox->playlist.cur == -1)) {
	      printf("tox->playlist.num && (tox->playlist.cur == -1)\n");
	      tox->playlist.cur = 0;
	      toxine_set_current_mrl_from_cur(tox);
	    }
	  }

	  i++;
	}
      }
    }
    else if(toxine_is_arg_contain(tox, 1, "load")) {

      playlist_load(tox, (toxine_get_arg(tox, 2)));
      
      tox->playlist.loop = 0;

      if(tox->playlist.num) {
	tox->playlist.cur = 0;
	toxine_set_current_mrl_from_cur(tox);
      }
    }
    else if(toxine_is_arg_contain(tox, 1, "save"))
      playlist_save(tox, (toxine_get_arg(tox, 2)));
    else if(toxine_is_arg_contain(tox, 1, "loop")) {

      if(toxine_is_arg_contain(tox, 2, "status"))
	pinfo("Playlist loop mode: %s [%s]\n",
	      ((tox->playlist.loop) ? "ON" : "OFF"), loop_modes[tox->loop_mode].name);
      else
	tox->playlist.loop = (toxine_get_bool_value((toxine_get_arg(tox, 2)))) ? 1 : 0;

    }
  }
}

static void do_display(commands_t *command, toxine_t *tox, void *data) {
  int nargs;

  nargs = toxine_is_args(tox);

  if(nargs == 1) {
    if(toxine_is_arg_contain(tox, 1, "settings")) {
      
      pinfo("Current settings:\n");
      pinfo("-----------------\n");
      pinfo("\n");
      pinfo("  Xine engine:\n");
      pinfo("    initialized: '%s'\n", (tox->xine) ? "YES" : "NO");
      pinfo("    display events: '%s'\n", (tox->display_events) ? "YES" : "NO");
      pinfo("\n");
      pinfo("  Video:\n");
      pinfo("    number of toxine plugins: '%d'\n", tox->video.plugins_num);
      pinfo("    driver name: '%s'\n", (tox->video.name) ? tox->video.name : "not set");
      pinfo("    spu channel: '%d'\n", tox->video.spu_channel);
      pinfo("\n");
      pinfo("  Audio:\n");
      pinfo("    driver name: '%s'\n", (tox->audio.name) ? tox->audio.name : "not set");
      pinfo("    capabilities: '0x%x\n", tox->audio.mixer.caps);
      pinfo("    mixer enabled: '%s'\n", (tox->audio.mixer.enable) ? "YES" : "NO");
      pinfo("    channel: '%d'\n", tox->audio.channel);
      pinfo("    volume: '%d'\n", tox->audio.mixer.volume_level);
      pinfo("    mute: '%s'\n", (tox->audio.mixer.mute) ? "ON" : "OFF");
      pinfo("\n");
      pinfo("  Toxine:\n");
      pinfo("    interactive mode: '%s'\n", (tox->interactive) ? "YES" : "NO");
      pinfo("    mrl: '%s'\n", (tox->current_mrl) ? tox->current_mrl : "not set");
      pinfo("    ignore finished: '%d'\n", tox->ignore_finished);
      pinfo("    Playlist/Mediamark:\n");
      pinfo("      number of entries: '%d'\n", tox->playlist.num);
      pinfo("      current: '%d'\n", tox->playlist.cur);
      pinfo("      loop: '%s' [%s]\n", 
	    (tox->playlist.loop) ? "YES" : "NO", loop_modes[tox->loop_mode].name);
      pinfo("    Script:\n");
      pinfo("      in use: '%s'\n", (tox->script.in_use) ? "YES" : "NO");
      pinfo("      file: '%s'\n", (tox->script.filename) ? tox->script.filename : "not set");
      pinfo(".\n");

    }
  }
  else if(nargs >= 2) {
    if(toxine_is_arg_contain(tox, 1, "current")) {
      if(toxine_is_arg_contain(tox, 2, "position")) {
	int pos;

	return_if_no_init(tox);

	xine_get_pos_length(tox->stream, &pos, NULL, NULL);
	toxine_draw_bar("Current position", 0, 65535, pos);
      }
      else if(toxine_is_arg_contain(tox, 2, "time")) {
	int seconds;
	
	return_if_no_init(tox);
	
	xine_get_pos_length(tox->stream, NULL, &seconds, NULL);
	seconds /= 1000;
	pinfo("Current time: %02d:%02d:%02d\n",
	      seconds / (60 * 60), (seconds / 60) % 60, seconds % 60);
	
      }
    }
    else if(toxine_is_arg_contain(tox, 1, "events")) {
      const char *arg = toxine_get_arg(tox, 2);

      if(toxine_is_arg_contain(tox, 2, "status")) {
	pinfo("Display events: %s\n", (tox->display_events) ? "YES" : "NO");
      }
      else
	tox->display_events = toxine_get_bool_value(arg);
    }
  }
}

static void do_mediamk(commands_t *command, toxine_t *tox, void *data) {
  int nargs;

  nargs = toxine_is_args(tox);

  if(nargs >= 1 && (toxine_is_arg_contain(tox, 1, "add"))) {
    mediamark_t   mmk;
    
    if(nargs == 1) {
      int           confirm;
      int           question_counter = 0;
      char         *line;
      char         *questions[] = {
	"      Mrl ?: ",
	"      Identificator (empty: use mrl)?: ",
	"      Start time in secs (empty for stream start)?: ",
	"      End time in secs (empty for stream end)?: ",
	NULL
      };
      
      while(questions[question_counter] != NULL) {
	
      __re_readline_1:
	if((line = readline(questions[question_counter])) == NULL) {
	  perr("%s(%d): readline() failed: %s\n", __tox_func__, __LINE__,
	       strerror(errno));
	  goto __re_readline_1;
	}
	
	switch(question_counter) {
	case 0:
	  if(*line) {
	    char *expanded_mrl;
	    
	    expanded_mrl = tilde_expand((toxine_atoa(line)));
	    xine_strdupa(mmk.mrl, expanded_mrl);
	    free(expanded_mrl);
	  }
	  else {
	    perr("Mrl should be specified, at least.\n");
	    perr("Action canceled.\n");
	    return;
	  }
	  break;
	case 1:
	  if(*line)
	    xine_strdupa(mmk.ident, (toxine_atoa(line)));
	  else
	    xine_strdupa(mmk.ident, mmk.mrl);
	  break;
	case 2:
	  if(*line)
	    mmk.start = atoi(line);
	  else
	    mmk.start = 0;
	  break;
	case 3:
	  if(*line)
	    mmk.end = atoi(line);
	  else
	    mmk.end = -1;
	  break;
	}
	
	question_counter++;
      }
      
      pinfo("New mediamark:\n");
      pinfo("  ident: '%s'\n", mmk.ident);
      pinfo("  mrl: '%s'\n", mmk.mrl);
      pinfo("  start @ %d, end @ %d.\n", mmk.start, mmk.end);
      confirm = (tox->interactive) ? (toxine_confirm("Add mediamark ? [yes|no]: ")) : 1;
      
      if(confirm) {
	playlist_add_entry(tox, mmk.mrl, mmk.ident, mmk.start, mmk.end);
	
	if(tox->playlist.cur == -1) {
	  tox->playlist.cur = 0;
	  toxine_set_current_mrl_from_cur(tox);
	}
	
      }
      else {
	poutalign();
	pout("Canceled.\n");
      }	  
      
    }
    else {
      int     i = 2;
      int     mmk_members[4];
      
      mmk_members[0] = 0;  /* ident */
      mmk_members[1] = -1; /* mrl */
      mmk_members[2] = 0;  /* start */
      mmk_members[3] = 0;  /* end */

      mmk.start = 0;
      mmk.end = -1;
      
      if(nargs % 2) {
	
	for(i = 2; i < nargs; i += 2) {
	  
	  if((toxine_is_arg_contain(tox, i, "ident")) && (mmk_members[0] == 0)) {
	    mmk_members[0] = 1;
	    xine_strdupa(mmk.ident, ((char *)toxine_get_arg(tox, i + 1)));
	  }
	  else if((toxine_is_arg_contain(tox, i, "mrl")) && (mmk_members[1] == -1)) {
	    char *expanded_mrl;
	    
	    mmk_members[1] = 1;
	    expanded_mrl = tilde_expand(((char *)toxine_get_arg(tox, i + 1)));
	    xine_strdupa(mmk.mrl, expanded_mrl);
	    
	    free(expanded_mrl);
	  }
	  else if((toxine_is_arg_contain(tox, i, "start")) && (mmk_members[0] == 0)) {
	    mmk_members[2] = 1;
	    mmk.start = atoi(((char *)toxine_get_arg(tox, i + 1)));
	  }
	  else if((toxine_is_arg_contain(tox, i, "end")) && (mmk_members[0] == 0)) {
	    mmk_members[3] = 1;
	    mmk.end = atoi(((char *)toxine_get_arg(tox, i + 1)));
	  }
	  
	}
	
	if(mmk_members[1] != -1) {
	  int   confirm;

	  if(mmk_members[0] == 0)
	    xine_strdupa(mmk.ident, mmk.mrl);
	  
	  if((mmk.end != -1) && ((mmk.start >= mmk.end) || (mmk.end <= mmk.start))) {
	    perr("start and end time are wrong:start @ %d, end @ %d\n", mmk.start, mmk.end);
	    pinfo("Set start/end times to default\n");
	    mmk.start = 0;
	    mmk.end = -1;
	  }

	  pinfo("New mediamark:\n");
	  pinfo("  ident: '%s'\n", mmk.ident);
	  pinfo("  mrl: '%s'\n", mmk.mrl);
	  pinfo("  start @ %d, end @ %d.\n", mmk.start, mmk.end);
	  confirm = (tox->interactive) ? (toxine_confirm("Add mediamark ? [yes|no]: ")) : 1;
	  
	  if(confirm) {
	    playlist_add_entry(tox, mmk.mrl, mmk.ident, mmk.start, mmk.end);
	    
	    if(tox->playlist.cur == -1) {
	      tox->playlist.cur = 0;
	      toxine_set_current_mrl_from_cur(tox);
	    }

	  }
	  else {
	    poutalign();
	    pout("Canceled.\n");
	  }	  
	}
	else
	  perr("MRL wasn't specified\n");
	  
      }
      else {
	int i = 1;
	
	perr("Wrong number of arguments:\n");
	poutalign();
	pout("'");
	while(i <= nargs) {
	  pout("%s ", toxine_get_arg(tox, i));
	  i++;
	}
	pout("\b'\n");
      }

    }
  }
  else if(nargs == 1) {
    if(toxine_is_arg_contain(tox, 1, "show")) {
      
      if(tox->playlist.num) {
	int i;
	
	pinfo("Mediamark list (* is current):\n");
	for(i = 0; i < tox->playlist.num; i++) {

	  poutalign();

	  if(tox->playlist.cur == i)
	    pout("*%4d - ident: '%s'\n", i, tox->playlist.mmk[i]->ident);
	  else
	    pout(" %4d - ident:'%s'\n", i, tox->playlist.mmk[i]->ident);

	  poutalign();
	  pout("        mrl: '%s'\n", tox->playlist.mmk[i]->mrl);
	  poutalign();
	  pout("        start @ %d, end @ %d.\n", 
	       tox->playlist.mmk[i]->start, tox->playlist.mmk[i]->end);

	}

	pinfo(".\n");

      }
      else
	pinfo("Empty playlist.\n");

    }
  }
  else if(nargs >= 2) {

    if(toxine_is_arg_contain(tox, 1, "load")) {
      playlist_load(tox, (toxine_get_arg(tox, 2)));
      
      tox->playlist.loop = 0;
      
      if(tox->playlist.num) {
	tox->playlist.cur = 0;
	toxine_set_current_mrl_from_cur(tox);
      }

    }
    else if(toxine_is_arg_contain(tox, 1, "save"))
      playlist_save(tox, (toxine_get_arg(tox, 2)));
    else if(toxine_is_arg_contain(tox, 1, "delete")) {

      if(toxine_is_arg_contain(tox, 2, "all") || toxine_is_arg_contain(tox, 2, "*")) {
	if(tox->playlist.num) {
	  
	  for(; tox->playlist.num > 0;)
	    playlist_free_entry(tox, (tox->playlist.num - 1));
	  
	  toxine_free(tox->playlist.mmk);
	  
	  tox->playlist.num = 0;
	  tox->playlist.cur = -1;
	  tox->playlist.loop = 0;
	  
	  toxine_set_current_mrl_from_cur(tox);
	}
      }
      else {
	const char *argument;
	int         i = 2;
	
	while((argument = toxine_get_arg(tox, i))) {
	  int entry;
	  
	  if((entry = playlist_get_entry_from_id(tox, argument)) < 0)
	    entry = atoi(toxine_get_arg(tox, 2));
	  
	  if(tox->playlist.num && (entry < tox->playlist.num)) {
	    int j;
	    
	    if(tox->playlist.cur >= entry) {
	      tox->playlist.cur = -1;
	      tox->current_mrl = NULL;
	    }

	    playlist_free_entry(tox, entry);
	    
	    for(j = entry; j < tox->playlist.num; j++)
	      tox->playlist.mmk[j] = tox->playlist.mmk[j + 1];

	    tox->playlist.mmk = (mediamark_t **) 
	      realloc(tox->playlist.mmk, sizeof(mediamark_t *) * (tox->playlist.num + 2));
	    
	    tox->playlist.mmk[tox->playlist.num + 1] = NULL;
	    
	    if(!tox->playlist.num)
	      tox->playlist.loop = 0;
	    
	    if(tox->playlist.num && (tox->playlist.cur == -1)) {
	      tox->playlist.cur = 0;
	      toxine_set_current_mrl_from_cur(tox);
	    }
	  }

	  i++;
	}
      }

    }
    else if(toxine_is_arg_contain(tox, 1, "edit")) {
      int entry;
      
      if((entry = playlist_get_entry_from_id(tox, toxine_get_arg(tox, 2))) < 0)
	entry = atoi(toxine_get_arg(tox, 2));
      
      if(entry < tox->playlist.num) {
	char         *line;
	mediamark_t   mmk;
	int           question_counter = 0;
	int           changed = 0;
	char         *questions[] = {
	  "      Identificator (empty to keep original)?: ",
	  "      Mrl (empty to keep original)?: ",
	  "      Start time in secs (empty to keep original)?: ",
	  "      End time in secs (empty to keep original, -1 to stream end)?: ",
	  NULL
	};
	
	pinfo("Editing :\n");
	pinfo(" %4d - ident: '%s'\n", entry, tox->playlist.mmk[entry]->ident);
	pinfo("        mrl: '%s'\n", tox->playlist.mmk[entry]->mrl);
	pinfo("        start @ %d, end @ %d.\n", 
	      tox->playlist.mmk[entry]->start, tox->playlist.mmk[entry]->end);

	while(questions[question_counter] != NULL) {
	  
	__re_readline_2:
	  if((line = readline(questions[question_counter])) == NULL) {
	    perr("%s(%d): readline() failed: %s\n", __tox_func__, __LINE__,
		 strerror(errno));
	    goto __re_readline_2;
	  }
	  
	  if(*line)
	    changed++;
	  
	  switch(question_counter) {
	  case 0:
	    if(*line)
	      xine_strdupa(mmk.ident, (toxine_atoa(line)));
	    else
	      xine_strdupa(mmk.ident, tox->playlist.mmk[entry]->ident);
	    break;
	  case 1:
	    if(*line) {
	      char *expanded_mrl;

	      expanded_mrl = tilde_expand((toxine_atoa(line)));
	      xine_strdupa(mmk.mrl, (toxine_atoa(line)));
	      free(expanded_mrl);
	    }
	    else
	      xine_strdupa(mmk.mrl, tox->playlist.mmk[entry]->mrl);
	    break;
	  case 2:
	    if(*line)
	      mmk.start = atoi(line);
	    else
	      mmk.start = tox->playlist.mmk[entry]->start;
	    break;
	  case 3:
	    if(*line)
	      mmk.end = atoi(line);
	    else
	      mmk.end = tox->playlist.mmk[entry]->end;
	    break;
	  }
	  
	  question_counter++;
	}
	
	/* Summary */
	if(changed) {
	  int confirm;

	  pinfo("New mediamark:\n");
	  pinfo(" %4d - ident: '%s'\n", entry, mmk.ident);
	  pinfo("        mrl: '%s'\n", mmk.mrl);
	  pinfo("        start @ %d, end @ %d.\n", mmk.start, mmk.end);

	  confirm = (tox->interactive) ? (toxine_confirm("Replace mediamark ? [yes|no]: ")) : 1;
	  
	  if(confirm) {
	    playlist_replace_entry(&tox->playlist.mmk[entry], 
				   mmk.mrl, mmk.ident, mmk.start, mmk.end);
	    pinfo("Mediamark updated.\n");
	  }
	  else {
	    poutalign();
	    pout("Canceled.\n");
	  }
	}
	else
	  pinfo("Nothing has been changed for this mediamark.\n");

      }
      else
	perr("Entry %d is out of bounds\n", entry);

    }
    else if(toxine_is_arg_contain(tox, 1, "select")) {
      int entry;
      int old_cur = tox->playlist.cur;
      
      if((entry = playlist_get_entry_from_id(tox, toxine_get_arg(tox, 2))) < 0)
	entry = atoi(toxine_get_arg(tox, 2));
      
      if(entry < tox->playlist.num) {
	tox->playlist.cur = entry;
	toxine_set_current_mrl_from_cur(tox);
      }
      
      /*
       * If xine engine is in PLAY status, and new selected entry is != than old one, 
       * start playback with new selected stream 
       */
      if(tox->xine && (old_cur != tox->playlist.cur)) {
	int status = xine_get_status(tox->stream);
	
	if(status == XINE_STATUS_PLAY)
	  _xine_play(NULL, tox, (void *)NO_ARGS);
      }
    }
    else if(toxine_is_arg_contain(tox, 1, "loop")) {
      
      if(toxine_is_arg_contain(tox, 2, "status"))
	pinfo("Mediamark loop mode: %s\n", ((tox->playlist.loop) ? "ON" : "OFF"));
      else
	tox->playlist.loop = (toxine_get_bool_value((toxine_get_arg(tox, 2)))) ? 1 : 0;

    }
  }
}

static char *get_var(toxine_t *tox, const char *var) {

  if(tox && var) {
    char  *variable;
    
    variable = strchr(var, '$');

    if(variable && (*variable == '$') && (*(variable + 1) != '$')) {
      static char   buffer[32768];
      char          varname[1024];
      int           found = 0;
      
      memset(buffer, 0, sizeof(buffer));
      memset(varname, 0, sizeof(varname));

      if(sscanf(variable, "$\(%[a-z-A-Z-0-9-_])", &varname[0]) == 1)
	found = 1;
      else if(sscanf(variable, "$\{%[a-z-A-Z-0-9-_]}", &varname[0]) == 1)
	found = 1;
      else if(sscanf(variable, "$%[a-z-A-Z-0-9-_]", &varname[0]) == 1)
	found = 1;
      
      if(found) {
	static const char day_names[] = "SunMonTueWedThuFriSat";
	static const char month_names[] = "JanFebMarAprMayJunJulAugSepOctNovDec";
	
	found = 0;
	if(!strncasecmp(varname, "result", 6)) {
	  snprintf(buffer, sizeof(buffer), "%d", toxine_get_last_int_result(tox));
	  found = 1;
	}
	else if(!strncasecmp(varname, "cresult", 7)) {
	  snprintf(buffer, sizeof(buffer), "%s", toxine_get_last_char_result(tox));
	  found = 1;
	}
	else if(!strncasecmp(varname, "cxresult", 7)) {
	  char *p = toxine_get_last_char_result(tox);
	  if(p) {
	    int val = strtol(p, &p, 10);
	    snprintf(buffer, sizeof(buffer), "0x%x", val);
	  }
	  else
	    snprintf(buffer, sizeof(buffer), "Nothing available");
	  found = 1;
	}
	else if(!strncasecmp(varname, "iresult", 7)) {
	  snprintf(buffer, sizeof(buffer), "%d", toxine_get_last_int_result(tox));
	  found = 1;
	}
	else if(!strncasecmp(varname, "xresult", 7)) {
	  snprintf(buffer, sizeof(buffer), "0x%x", toxine_get_last_int_result(tox));
	  found = 1;
	}
	else if(!strncasecmp(varname, "version", 7)) {
	  snprintf(buffer, sizeof(buffer), "%s", VERSION);
	  found = 1;
	}
	else if(!strncasecmp(varname, "date", 4)) {
	  time_t      curtime;
	  struct tm  *lt;
	  
	  time(&curtime);
	  lt = localtime(&curtime);
	  snprintf(buffer, sizeof(buffer), "%-3.3s %-3.3s %d %d",
		   day_names + 3 * lt->tm_wday, month_names + 3 * lt->tm_mon, 
		   lt->tm_mday, (lt->tm_year + 1900));
	  found = 1;
	}
	else if(!strncasecmp(varname, "time", 4)) {
	  time_t      curtime;
	  struct tm  *lt;
	  
	  time(&curtime);
	  lt = localtime(&curtime);
	  snprintf(buffer, sizeof(buffer), "%02d:%02d:%02d", lt->tm_hour, lt->tm_min, lt->tm_sec);
	  found = 1;
	}
	else if(!strncasecmp(varname, "config_file", 11)) {
	  snprintf(buffer, sizeof(buffer), "%s", (tox->configfile) ? tox->configfile : "UNSET");
	  found = 1;
	}
	else if(!strncasecmp(varname, "video_driver", 12)) {
	  snprintf(buffer, sizeof(buffer), "%s", (tox->video.name) ? tox->video.name : "UNSET");
	  found = 1;
	}
	else if(!strncasecmp(varname, "audio_driver", 12)) {
	  snprintf(buffer, sizeof(buffer), "%s", (tox->audio.name) ? tox->audio.name : "UNSET");
	  found = 1;
	}
	else if(!strncasecmp(varname, "mrl", 3)) {
	  snprintf(buffer, sizeof(buffer), "%s", (tox->current_mrl) ? tox->current_mrl : "UNSET");
	  found = 1;
	}
	else if(!strncasecmp(varname, "script_file", 11)) {
	  snprintf(buffer, sizeof(buffer), "%s", (tox->script.filename) ? tox->script.filename : "UNSET");
	  found = 1;
	}

	if(found)
	  return &buffer[0];

      }
    }    
  }

  return NULL;
}

static void do_echo(commands_t *command, toxine_t *tox, void *data) {
  int         i;
  int         nargs = toxine_is_args(tox);
  int         len = 0;
  const char *var, *arg;

  i = 1;
  while(i <= nargs) {
    arg = toxine_get_arg(tox, i++);
    if((var = get_var(tox, arg)) != NULL)
      len += strlen(var);
    else
      len += strlen(arg);
  }
  
  if(len) {
    len += nargs;
    
    {
      char buffer[(len * 2)];
      
      memset(&buffer, 0, sizeof(buffer));
      
      i = 1;
      while(i <= nargs) {

	if(i > 1)
	  strcat(buffer, " ");

	arg = toxine_get_arg(tox, i++);

	if((var = get_var(tox, arg)) != NULL)
	  strcat(buffer, var);
	else
	  strcat(buffer, arg);

      }
      sprintf(buffer, "%s%c", buffer, '\n');
      
      pecho(buffer);
    }
  }

}

static void do_dumpconfig(commands_t *command, toxine_t *tox, void *data) {
  _cfg_entry_t  cfg_entry;

  return_if_no_new(tox);

  pinfo("Dumping all config object:\n");
  if((xine_config_get_first_entry(tox->xine, &cfg_entry))) {
    do {
      toxine_dump_config_entry(tox, &cfg_entry);
    } while((xine_config_get_next_entry(tox->xine, &cfg_entry)));
  }
  pinfo(".\n");

  pinfo("\n");
}

static void do_dumpstream(commands_t *command, toxine_t *tox, void *data) {
  _NAPI_xine_dumpstream(tox);
}
