/*
** Copyright (C) 2002-2004 Daniel Caujolle-Bert <segfault@club-internet.fr>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>

#include "common.h"

typedef struct {
  int     init;
} dxr3_private_t;


static void init_video_out(toxine_t *tox) {
  toxine_vo_plugin_t *vop = (toxine_vo_plugin_t *) tox->video.cur_plugin;
  dxr3_private_t     *private = (dxr3_private_t *) vop->private;
  char               *drivername = "dxr3";
  int                 mode = 0;
  static char       *videoout_modes[] = { 
    "letterboxed tv", 
    "widescreen tv", 
    "overlay",
    NULL 
  };
  
  if(!strcasecmp(tox->video.name, "dxr3_lb"))
    mode = 0;
  else if(!strcasecmp(tox->video.name, "dxr3_ws"))
    mode = 1;

  /* First, we have to initialize the config entry */
  (void *) xine_config_register_enum(tox->xine, "dxr3.videoout_mode", 0, videoout_modes,
				     "Dxr3: videoout mode (tv or overlay)", 
				      NULL, 0, NULL, NULL);
  
  /* Then, force it to TV mode */
  toxine_config_update_num(tox, "dxr3.videoout_mode", mode);
  /* Force saving */
  toxine_config_save(tox);
  /* Reload config file now */
  toxine_config_load(tox);
  
  tox->video.port = xine_open_video_driver(tox->xine, drivername, XINE_VISUAL_TYPE_X11, NULL);

  private->init = 1;
  
  if(tox->video.port == NULL) {
    fprintf(stderr, "xine_load_video_output_plugin() failed to load '%s' driver.\n",
	    drivername);
    exit(1);
  }
  else
    pinfo("video driver '%s' successfully loaded.\n", drivername);
  
}

static void deinit_video_out(toxine_t *tox) {
  toxine_vo_plugin_t *vop = (toxine_vo_plugin_t *) tox->video.cur_plugin;
  dxr3_private_t     *private = (dxr3_private_t *) vop->private;
 
  xine_close_video_driver(tox->xine, tox->video.port);
  tox->video.port = NULL;
  
  private->init = 0;
}

static void release_video_out(toxine_t *tox) {
  toxine_vo_plugin_t *vop = (toxine_vo_plugin_t *) tox->video.last_plugin;
  dxr3_private_t     *private = (dxr3_private_t *) vop->private;

  if(private->init)
    deinit_video_out(tox);

  free(private);
  free(vop);
}

static const char *get_identifier(void) {
  return "DXR3/HW+ TV out";
}

static char *_vo_dxr3_names[] = { 
  "dxr3_lb", "dxr3_ws", 
  NULL 
};
static const char **get_names(void) {
  return (const char **)_vo_dxr3_names;
}

static const char *get_help(void) {
  return "Video out plugin supports DXR3/HW+ cards (lb:letterboxed, ws:widescreen).\n";
}

static int is_fullscreen(toxine_t *tox) {
  return 0;
}

static void fullscreen(toxine_t *tox) {
}

toxine_vo_plugin_t *toxine_load_vo_plugin(toxine_t *tox) {
  toxine_vo_plugin_t  *dxr3;
  dxr3_private_t      *private;
  
  dxr3 = (toxine_vo_plugin_t *) xine_xmalloc(sizeof(toxine_vo_plugin_t));
  private = (dxr3_private_t *) xine_xmalloc(sizeof(dxr3_private_t));

  private->init              = 0;

  dxr3->video_out_init       = init_video_out;
  dxr3->video_out_event_loop = NULL;
  dxr3->video_out_deinit     = deinit_video_out;
  dxr3->video_out_release    = release_video_out;
  dxr3->get_identifier       = get_identifier;
  dxr3->get_names            = get_names;
  dxr3->get_help             = get_help;
  dxr3->receive_xine_event   = NULL;
  dxr3->is_fullscreen        = is_fullscreen;
  dxr3->fullscreen           = fullscreen;
  dxr3->private              = (void *) private;

  return dxr3;
}
